@@ -1,6 +1,6 @@
 # imagecodecs/numcodecs.py
 
-# Copyright (c) 2021-2023, Christoph Gohlke
+# Copyright (c) 2021-2024, Christoph Gohlke
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -35,18 +35,18 @@ from __future__ import annotations
 
 __all__ = ['register_codecs']
 
-import numpy
-from numcodecs.abc import Codec
-from numcodecs.registry import register_codec, get_codec
+from typing import TYPE_CHECKING
 
 import imagecodecs
-
-from typing import TYPE_CHECKING
+import numpy
+from numcodecs.abc import Codec
+from numcodecs.registry import get_codec, register_codec
 
 if TYPE_CHECKING:
     from collections.abc import Sequence
     from typing import Any, Literal
-    from numpy.typing import NDArray
+
+    from numpy.typing import DTypeLike, NDArray
 
 
 class Aec(Codec):
@@ -175,7 +175,9 @@ class Avif(Codec):
         )
 
     def decode(self, buf, out=None):
-        return imagecodecs.avif_decode(buf, index=self.index, out=out)
+        return imagecodecs.avif_decode(
+            buf, index=self.index, numthreads=self.numthreads, out=out
+        )
 
 
 class Bitorder(Codec):
@@ -317,6 +319,33 @@ class Blosc2(Codec):
         )
 
 
+class Bmp(Codec):
+    """BMP codec for numcodecs."""
+
+    codec_id = 'imagecodecs_bmp'
+
+    def __init__(
+        self,
+        *,
+        ppm: int | None = None,
+        asrgb: bool | None = None,
+        squeeze: Literal[False] | Sequence[int] | None = None,
+    ) -> None:
+        if not imagecodecs.BMP.available:
+            raise ValueError('imagecodecs.BMP not available')
+
+        self.ppm = None if ppm is None else max(1, int(ppm))
+        self.asrgb = None if asrgb is None else bool(asrgb)
+        self.squeeze = squeeze
+
+    def encode(self, buf):
+        buf = _image(buf, self.squeeze)
+        return imagecodecs.bmp_encode(buf, ppm=self.ppm)
+
+    def decode(self, buf, out=None):
+        return imagecodecs.bmp_decode(buf, asrgb=self.asrgb, out=out)
+
+
 class Brotli(Codec):
     """Brotli codec for numcodecs."""
 
@@ -354,7 +383,7 @@ class Byteshuffle(Codec):
         self,
         *,
         shape: tuple[int, ...],
-        dtype: numpy.dtype | str,
+        dtype: DTypeLike,
         axis: int = -1,
         dist: int = 1,
         delta: bool = False,
@@ -372,8 +401,10 @@ class Byteshuffle(Codec):
 
     def encode(self, buf):
         buf = numpy.asarray(buf)
-        assert buf.shape == self.shape
-        assert buf.dtype == self.dtype
+        if buf.shape != self.shape:
+            raise ValueError(f'{buf.shape=} does not match {self.shape=}')
+        if buf.dtype != self.dtype:
+            raise ValueError(f'{buf.dtype=} does not match {self.dtype=}')
         return imagecodecs.byteshuffle_encode(
             buf,
             axis=self.axis,
@@ -472,7 +503,7 @@ class Checksum(Codec):
             if prepend is None:
                 prepend = False
         else:
-            raise ValueError(f'checksum kind {kind!r} not supported')
+            raise ValueError(f'checksum {kind=!r} not supported')
 
         self.kind = kind
         self.value = value
@@ -589,7 +620,7 @@ class Delta(Codec):
         self,
         *,
         shape: tuple[int, ...] | None = None,
-        dtype: numpy.dtype | str | None = None,
+        dtype: DTypeLike = None,
         axis: int = -1,
         dist: int = 1,
     ) -> None:
@@ -604,8 +635,10 @@ class Delta(Codec):
     def encode(self, buf):
         if self.shape is not None or self.dtype is not None:
             buf = numpy.asarray(buf)
-            assert buf.shape == self.shape
-            assert buf.dtype == self.dtype
+            if buf.shape != self.shape:
+                raise ValueError(f'{buf.shape=} does not match {self.shape=}')
+            if buf.dtype != self.dtype:
+                raise ValueError(f'{buf.dtype=} does not match {self.dtype=}')
         return imagecodecs.delta_encode(
             buf, axis=self.axis, dist=self.dist
         ).tobytes()
@@ -620,6 +653,25 @@ class Delta(Codec):
         )
 
 
+class Dicomrle(Codec):
+    """DICOMRLE codec for numcodecs."""
+
+    codec_id = 'imagecodecs_dicomrle'
+
+    def __init__(self, *, dtype: DTypeLike) -> None:
+        if not imagecodecs.DICOMRLE.available:
+            raise ValueError('imagecodecs.DICOMRLE not available')
+
+        self.dtype = numpy.dtype(dtype).str  # TODO: preserve endianness
+
+    def encode(self, buf, out=None):
+        # buf = _image(buf, self.squeeze)
+        raise NotImplementedError
+
+    def decode(self, buf, out=None):
+        return imagecodecs.dicomrle_decode(buf, self.dtype, out=out)
+
+
 class Eer(Codec):
     """Electron Event Representation codec for numcodecs."""
 
@@ -665,7 +717,7 @@ class Float24(Codec):
 
     def __init__(
         self,
-        byteorder: Literal['>'] | Literal['<'] | Literal['='] | None = None,
+        byteorder: Literal['>', '<', '='] | None = None,
         rounding: int | None = None,
     ) -> None:
         if not imagecodecs.FLOAT24.available:
@@ -695,7 +747,7 @@ class Floatpred(Codec):
         self,
         *,
         shape: tuple[int, ...],
-        dtype: numpy.dtype | str,
+        dtype: DTypeLike,
         axis: int = -1,
         dist: int = 1,
     ) -> None:
@@ -709,8 +761,10 @@ class Floatpred(Codec):
 
     def encode(self, buf):
         buf = numpy.asarray(buf)
-        assert buf.shape == self.shape
-        assert buf.dtype == self.dtype
+        if buf.shape != self.shape:
+            raise ValueError(f'{buf.shape=} does not match {self.shape=}')
+        if buf.dtype != self.dtype:
+            raise ValueError(f'{buf.dtype=} does not match {self.dtype=}')
         return imagecodecs.floatpred_encode(
             buf, axis=self.axis, dist=self.dist
         ).tobytes()
@@ -1122,7 +1176,7 @@ class Lerc(Codec):
         # masks: ArrayLike | None = None,
         version: int | None = None,
         planar: bool | None = None,
-        compression: Literal['zstd'] | Literal['deflate'] | None = None,
+        compression: Literal['zstd', 'deflate'] | None = None,
         compressionargs: dict[str, Any] | None = None,
         squeeze: Literal[False] | Sequence[int] | None = None,
     ) -> None:
@@ -1242,6 +1296,32 @@ class Lz4f(Codec):
         return imagecodecs.lz4f_decode(buf, out=_flat(out))
 
 
+class Lz4h5(Codec):
+    """LZ4H5 codec for numcodecs."""
+
+    codec_id = 'imagecodecs_lz4h5'
+
+    def __init__(
+        self,
+        *,
+        level: int | None = None,
+        blocksize: int | None = None,
+    ) -> None:
+        if not imagecodecs.LZ4H5.available:
+            raise ValueError('imagecodecs.LZ4H5 not available')
+
+        self.level = level
+        self.blocksize = blocksize
+
+    def encode(self, buf):
+        return imagecodecs.lz4h5_encode(
+            buf, level=self.level, blocksize=self.blocksize
+        )
+
+    def decode(self, buf, out=None):
+        return imagecodecs.lz4h5_decode(buf, out=_flat(out))
+
+
 class Lzf(Codec):
     """LZF codec for numcodecs."""
 
@@ -1326,6 +1406,24 @@ class Lzma(Codec):
         return imagecodecs.lzma_decode(buf, out=_flat(out))
 
 
+class Lzo(Codec):
+    """LZO codec for numcodecs."""
+
+    codec_id = 'imagecodecs_lzo'
+
+    def __init__(self, *, header: bool = False) -> None:
+        if not imagecodecs.LZO.available:
+            raise ValueError('imagecodecs.LZO not available')
+        self.header = bool(header)
+
+    def encode(self, buf):
+        # return imagecodecs.lzo_encode(buf)
+        raise NotImplementedError
+
+    def decode(self, buf, out=None):
+        return imagecodecs.lzo_decode(buf, header=self.header, out=_flat(out))
+
+
 class Lzw(Codec):
     """LZW codec for numcodecs."""
 
@@ -1366,6 +1464,62 @@ class Packbits(Codec):
         return imagecodecs.packbits_decode(buf, out=_flat(out))
 
 
+class Packints(Codec):
+    """Packed integer codec for numcodecs."""
+
+    codec_id = 'imagecodecs_packints'
+
+    def __init__(
+        self, *, dtype: DTypeLike, bitspersample: int, runlen: int = 0
+    ) -> None:
+        if not imagecodecs.PACKINTS.available:
+            raise ValueError('imagecodecs.PACKINTS not available')
+
+        self.dtype = numpy.dtype(dtype).str
+        self.bitspersample = bitspersample
+        self.runlen = runlen
+
+    def encode(self, buf):
+        raise NotImplementedError
+
+    def decode(self, buf, out=None):
+        return imagecodecs.packints_decode(
+            buf,
+            dtype=self.dtype,
+            bitspersample=self.bitspersample,
+            runlen=self.runlen,
+            out=_flat(out),
+        )
+
+
+class Pcodec(Codec):
+    """Pcodec codec for numcodecs."""
+
+    codec_id = 'imagecodecs_pcodec'
+
+    def __init__(
+        self,
+        *,
+        shape: tuple[int, ...],
+        dtype: DTypeLike,
+        level: int | None = None,
+    ) -> None:
+        if not imagecodecs.PCODEC.available:
+            raise ValueError('imagecodecs.PCODEC not available')
+
+        self.shape = tuple(shape)
+        self.dtype = numpy.dtype(dtype).str
+        self.level = level
+
+    def encode(self, buf):
+        return imagecodecs.pcodec_encode(buf, level=self.level)
+
+    def decode(self, buf, out=None):
+        return imagecodecs.pcodec_decode(
+            buf, shape=self.shape, dtype=self.dtype, out=out
+        )
+
+
 class Pglz(Codec):
     """PGLZ codec for numcodecs."""
 
@@ -1456,6 +1610,31 @@ class Qoi(Codec):
         return imagecodecs.qoi_decode(buf, out=out)
 
 
+class Quantize(Codec):
+    """Quantize codec for numcodecs."""
+
+    codec_id = 'imagecodecs_quantize'
+
+    def __init__(
+        self,
+        *,
+        mode: Literal['bitgroom', 'granularbr', 'gbr', 'bitround', 'scale'],
+        nsd: int,  # number of significant digits
+    ) -> None:
+        if not imagecodecs.QUANTIZE.available:
+            raise ValueError('imagecodecs.QUANTIZE not available')
+
+        self.nsd = nsd
+        self.mode = mode
+
+    def encode(self, buf):
+        return imagecodecs.quantize_encode(buf, self.mode, self.nsd)
+
+    def decode(self, buf, out=None):
+        return buf
+        # return imagecodecs.quantize_decode(buf, self.mode, self.nsd, out=out)
+
+
 class Rcomp(Codec):
     """Rcomp codec for numcodecs."""
 
@@ -1465,7 +1644,7 @@ class Rcomp(Codec):
         self,
         *,
         shape: tuple[int, ...],
-        dtype: numpy.dtype | str,
+        dtype: DTypeLike,
         nblock: int | None = None,
     ) -> None:
         if not imagecodecs.RCOMP.available:
@@ -1542,6 +1721,70 @@ class Snappy(Codec):
         return imagecodecs.snappy_decode(buf, out=_flat(out))
 
 
+class Sperr(Codec):
+    """SPERR codec for numcodecs."""
+
+    codec_id = 'imagecodecs_sperr'
+
+    def __init__(
+        self,
+        *,
+        level: float,
+        mode: Literal['bpp', 'psnr', 'pwe'],
+        dtype: DTypeLike = None,
+        shape: tuple[int, ...] | None = None,
+        chunks: tuple[int, int, int] | None = None,
+        header: bool = True,
+        numthreads: int | None = None,
+        squeeze: Literal[False] | Sequence[int] | None = None,
+    ) -> None:
+        if not imagecodecs.SPERR.available:
+            raise ValueError('imagecodecs.SPERR not available')
+
+        if header:
+            self.shape = None
+            self.dtype = None
+        elif shape is None or dtype is None:
+            raise ValueError('invalid shape or dtype')
+        else:
+            self.shape = tuple(shape)
+            self.dtype = numpy.dtype(dtype).str
+        self.mode = mode
+        self.level = float(level)
+        self.chunks = None if chunks is None else tuple(chunks)
+        self.header = bool(header)
+        self.numthreads = numthreads
+        self.squeeze = squeeze
+
+    def encode(self, buf):
+        buf = _image(buf, self.squeeze)
+        if not self.header:
+            if buf.shape != self.shape:
+                raise ValueError(f'{buf.shape=} does not match {self.shape=}')
+            if buf.dtype != self.dtype:
+                raise ValueError(f'{buf.dtype=} does not match {self.dtype=}')
+        return imagecodecs.sperr_encode(
+            buf,
+            level=self.level,
+            mode=self.mode,
+            chunks=self.chunks,
+            header=self.header,
+            numthreads=self.numthreads,
+        )
+
+    def decode(self, buf, out=None):
+        if self.header:
+            return imagecodecs.sperr_decode(buf, out=out)
+        return imagecodecs.sperr_decode(
+            buf,
+            shape=self.shape,
+            dtype=numpy.dtype(self.dtype),
+            header=self.header,
+            numthreads=self.numthreads,
+            out=out,
+        )
+
+
 class Spng(Codec):
     """SPNG codec for numcodecs."""
 
@@ -1700,7 +1943,7 @@ class Xor(Codec):
         self,
         *,
         shape: tuple[int, ...] | None = None,
-        dtype: numpy.dtype | None = None,
+        dtype: DTypeLike = None,
         axis: int = -1,
     ) -> None:
         if not imagecodecs.XOR.available:
@@ -1713,8 +1956,10 @@ class Xor(Codec):
     def encode(self, buf):
         if self.shape is not None or self.dtype is not None:
             buf = numpy.asarray(buf)
-            assert buf.shape == self.shape
-            assert buf.dtype == self.dtype
+            if buf.shape != self.shape:
+                raise ValueError(f'{buf.shape=} does not match {self.shape=}')
+            if buf.dtype != self.dtype:
+                raise ValueError(f'{buf.dtype=} does not match {self.dtype=}')
         return imagecodecs.xor_encode(buf, axis=self.axis).tobytes()
 
     def decode(self, buf, out=None):
@@ -1734,7 +1979,7 @@ class Zfp(Codec):
         self,
         *,
         shape: tuple[int, ...] | None = None,
-        dtype: numpy.dtype | None = None,
+        dtype: DTypeLike = None,
         strides: tuple[int, ...] | None = None,
         level: int | None = None,
         mode: int | str | None = None,
@@ -1766,8 +2011,10 @@ class Zfp(Codec):
     def encode(self, buf):
         buf = numpy.asarray(buf)
         if not self.header:
-            assert buf.shape == self.shape
-            assert buf.dtype == self.dtype
+            if buf.shape != self.shape:
+                raise ValueError(f'{buf.shape=} does not match {self.shape=}')
+            if buf.dtype != self.dtype:
+                raise ValueError(f'{buf.dtype=} does not match {self.dtype=}')
         return imagecodecs.zfp_encode(
             buf,
             level=self.level,
